/*
 * main.c - VICE main startup entry.
 *
 * Written by
 *  Ettore Perazzoli <ettore@comm2000.it>
 *  Teemu Rantanen <tvr@cs.hut.fi>
 *  Vesa-Matti Puro <vmp@lut.fi>
 *  Jarkko Sonninen <sonninen@lut.fi>
 *  Jouko Valta <jopi@stekt.oulu.fi>
 *  Andre Fachat <a.fachat@physik.tu-chemnitz.de>
 *  Andreas Boose <viceteam@t-online.de>
 *
 * This file is part of VICE, the Versatile Commodore Emulator.
 * See README for copyright notice.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307  USA.
 *
 */

/* #define DEBUG_MAIN */

#include "vice.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <unistd.h>

#include "autostart.h"
#include "archdep.h"
//#include "cmdline.h"
#include "console.h"
#include "debug.h"
#include "drive.h"
#include "fullscreen.h"
//#include "gfxoutput.h"
#include "info.h"
#include "init.h"
//#include "initcmdline.h"
#include "lib.h"
#include "log.h"
#include "machine.h"
#include "maincpu.h"
#include "main.h"
#include "resources.h"
#include "sysfile.h"
#include "types.h"
#include "uiapi.h"
#include "uimsgbox.h"
#include "vsync.h"
#include "version.h"
#include "video.h"
#include "3ds.h"
#include "vice3ds.h"
#include "archdep_xdg.h"

#ifdef USE_SVN_REVISION
#include "svnversion.h"
#endif

#ifdef DEBUG_MAIN
#define DBG(x)  printf x
#else
#define DBG(x)
#endif

#ifdef __OS2__
const
#endif
int console_mode = 0;
int video_disabled_mode = 0;


/** \brief  Size of buffer used to write core team members' names to log/stdout
 *
 * 79 characters + 1 byte for '\0'. Assuming a terminal width of 80 characters,
 * we can only use 79 when calling log_message() since that function adds a
 * newline to its ouput.
 */
#define TERM_TMP_SIZE  80

static char *check_autostart_file(char *dir) {
	struct dirent *pDirent;
	DIR *pDir;

	pDir = opendir (dir);
	if (pDir == NULL) {
		archdep_startup_log_error("check autostart: cannot open directory %s\n", dir);
		return NULL;
	} else {
		while ((pDirent = readdir(pDir)) != NULL) {
			if (strncmp("autostart.",pDirent->d_name,10)==0) {
				char *p=archdep_join_paths(dir,pDirent->d_name,NULL);
				closedir(pDir);
				return p;
			}
		}
	}
	closedir (pDir);
	return NULL;
}

/* ------------------------------------------------------------------------- */
/* This is the main program entry point.  Call this from `main()'.  */
int main_program(int argc, char **argv)
{
    int ishelp = 0;

    // set 804 MHz
	if (isN3DS()) osSetSpeedupEnable(1);
	
	lib_init_rand();

    DBG(("main:archdep_init(argc:%d)\n", argc));
    if (archdep_init(&argc, argv) != 0) {
        archdep_startup_log_error("archdep_init failed.\n");
        return -1;
    }

	if (log_init() < 0) {
        archdep_startup_log_error("Cannot startup logging system.\n");
    }

    if (archdep_vice_atexit(main_exit) != 0) {
        archdep_startup_log_error("archdep_vice_atexit failed.\n");
        return -1;
    }

    maincpu_early_init();
    machine_setup_context();
    drive_setup_context();
    machine_early_init();

    /* Initialize system file locator.  */
    sysfile_init(machine_name);

    //gfxoutput_early_init(ishelp);
//    if ((init_resources() < 0) || (init_cmdline_options() < 0)) {
    if (init_resources() < 0) {
        return -1;
    }

    /* Set factory defaults.  */
    if (resources_set_defaults() < 0) {
        archdep_startup_log_error("Cannot set defaults.\n");
        return -1;
    }

    /* Initialize the user interface.  `ui_init()' might need to handle the
       command line somehow, so we call it before parsing the options.
       (e.g. under X11, the `-display' option is handled independently).  */
    DBG(("main:ui_init(argc:%d)\n", argc));
    if (!console_mode && ui_init(&argc, argv) < 0) {
        archdep_startup_log_error("Cannot initialize the UI.\n");
        return -1;
    }

    if (!ishelp) {
        /* Load the user's default configuration file.  */
        if (resources_load(NULL) < 0) {
            /* The resource file might contain errors, and thus certain
            resources might have been initialized anyway.  */
            if (resources_set_defaults() < 0) {
                archdep_startup_log_error("Cannot set defaults.\n");
                return -1;
            }
        }
    }

/*
    DBG(("main:initcmdline_check_args(argc:%d)\n", argc));
    if (initcmdline_check_args(argc, argv) < 0) {
        return -1;
    }
*/
	// 3DS - check for autostart file: first resource name AutostartImage, then in romfs and vice dir
	char *p;
	if ((resources_get_string("AutostartImage", (const char **)(&p)) == 0 && p != NULL && *p != 0 && access( p, F_OK ) != -1) ||
		(p=check_autostart_file("romfs:/"))!=NULL ||
		(p=check_autostart_file(archdep_xdg_data_home()))!=NULL) {
		autostart_string = p;
        autostart_mode = AUTOSTART_MODE_RUN;
	}

    /* VICE boot sequence.  */
    log_message(LOG_DEFAULT, " ");
    log_message(LOG_DEFAULT, "*** Vice3DS Version %s ***", VERSION3DS);
    log_message(LOG_DEFAULT, "*** based on VICE Version %s ***", VERSION);
    log_message(LOG_DEFAULT, " ");
    log_message(LOG_DEFAULT, "Welcome to Vice3DS, the free portable %s Emulator for Nintendo 3DS.", machine_name);
    log_message(LOG_DEFAULT, "Ported by badda71 <me@badda.de>");
    log_message(LOG_DEFAULT, " ");
    log_message(LOG_DEFAULT, "This is free software with ABSOLUTELY NO WARRANTY.");
    log_message(LOG_DEFAULT, "See the \"Help\" menu for more info.");
    log_message(LOG_DEFAULT, " ");

    /* Complete the GUI initialization (after loading the resources and
       parsing the command-line) if necessary.  */
    if (!console_mode && ui_init_finish() < 0) {
        return -1;
    }

    if (!console_mode && video_init() < 0) {
        return -1;
    }

/*    if (initcmdline_check_psid() < 0) {
        return -1;
    }
*/
    if (init_main() < 0) {
        return -1;
    }

	if (autostart_string != NULL) {
		autostart_autodetect_opt_prgname(autostart_string, 0, autostart_mode);
	}

    // initcmdline_check_attach();

	// check 3D-slider an warn the user if it is not set to zero
	if (slider3d_func && osGet3DSliderState() >= 0.1) {
		char *tmp = lib_msprintf("3D-Slider is not set to OFF. Emulation will run %s!",slider3d_func==1?"slower":"faster");
        message_box("VICE WARNING", tmp, MESSAGE_OK);
		free (tmp);
	}

    /* Let's go...  */
    log_message(LOG_DEFAULT, "Main CPU: starting at ($FFFC).");
    maincpu_mainloop();

    log_error(LOG_DEFAULT, "perkele!");

    return 0;
}
